import os

LOG_LEVEL = os.environ.get('LOG_LEVEL', 'INFO')
LOG_FORMAT = '%(asctime)s %(levelname)s:%(name)s:%(message)s'
LOG_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'

CONFIG_FILE = os.environ.get('CONFIG_FILE', 'config.yaml')

COMMON_SECTION_NAME = 'common'
TEMPLATES_DIR = os.environ.get('TEMPLATES_DIR', 'templates')

K8S_CONFIG_DIR = os.environ.get('K8S_CONFIG_DIR', '{}/.kube/'.format(os.path.expanduser('~')))

K8S_NAMESPACE = None

TEMP_DIR = os.environ.get('TEMP_DIR', '/tmp/k8s-handle')

CHECK_STATUS_TRIES = 360
CHECK_STATUS_TIMEOUT = 5

CHECK_CONTAINERS_IN_POD_TRIES = 360
CHECK_CONTAINERS_IN_POD_TIMEOUT = 5

CHECK_POD_STATUS_TRIES = 360
CHECK_POD_STATUS_TIMEOUT = 5

CHECK_DAEMONSET_STATUS_TRIES = 10
CHECK_DAEMONSET_STATUS_TIMEOUT = 5

COUNT_LOG_LINES = None

GET_ENVIRON_STRICT = False



import copy
import logging
import os
import re

from kubernetes import client

from k8s_handle import settings
from k8s_handle.dictionary import merge
from k8s_handle.filesystem import load_yaml, write_file_tmp
from k8s_handle.templating import b64decode

log = logging.getLogger(__name__)

INCLUDE_RE = re.compile(r'{{\s?file\s?=\s?\'(?P<file>[^\']*)\'\s?}}')
CUSTOM_ENV_RE = r'{{\s*env\s*=\s*\'([^\']*)\'\s*}}'

KEY_USE_KUBECONFIG = 'use_kubeconfig'
KEY_K8S_MASTER_URI = 'k8s_master_uri'
KEY_K8S_MASTER_URI_ENV = KEY_K8S_MASTER_URI.upper()
KEY_K8S_MASTER_URI_ENV_DEPRECATED = 'K8S_HOST'

KEY_K8S_CA_BASE64 = 'k8s_ca_base64'
KEY_K8S_CA_BASE64_ENV = KEY_K8S_CA_BASE64.upper()
KEY_K8S_CA_BASE64_URI_ENV_DEPRECATED = 'K8S_CA'

KEY_K8S_TOKEN = 'k8s_token'
KEY_K8S_TOKEN_ENV = KEY_K8S_TOKEN.upper()

KEY_K8S_NAMESPACE = 'k8s_namespace'
KEY_K8S_NAMESPACE_ENV = KEY_K8S_NAMESPACE.upper()
KEY_K8S_HANDLE_DEBUG = 'k8s_handle_debug'


class PriorityEvaluator:
    def __init__(self, cli_arguments, context_arguments, environment):
        self.cli_arguments = cli_arguments
        self.context_arguments = context_arguments
        self.environment = environment
        self.loaded = False

    def k8s_namespace_default(self, kubeconfig_namespace=None):
        return PriorityEvaluator._first(
            self.context_arguments.get(KEY_K8S_NAMESPACE),
            kubeconfig_namespace,
            self.environment.get(KEY_K8S_NAMESPACE_ENV))

    def k8s_client_configuration(self):
        for parameter, value in {
            KEY_K8S_MASTER_URI: self._k8s_master_uri(),
            KEY_K8S_CA_BASE64: self._k8s_ca_base64(),
            KEY_K8S_TOKEN: self._k8s_token()
        }.items():
            if value:
                continue

            raise RuntimeError(
                '{0} parameter is not set. Please, provide {0} via CLI, config or env.'.format(parameter))

        configuration = client.Configuration()
        configuration.host = self._k8s_master_uri()
        configuration.ssl_ca_cert = write_file_tmp(b64decode(self._k8s_ca_base64()).encode('utf-8'))
        configuration.api_key = {"authorization": "Bearer " + self._k8s_token()}
        configuration.debug = self._k8s_handle_debug()
        return configuration

    def environment_deprecated(self):
        return self.environment.get(KEY_K8S_MASTER_URI_ENV_DEPRECATED) or \
               self.environment.get(KEY_K8S_CA_BASE64_URI_ENV_DEPRECATED)

    def _k8s_master_uri(self):
        return PriorityEvaluator._first(
            self.cli_arguments.get(KEY_K8S_MASTER_URI),
            self.context_arguments.get(KEY_K8S_MASTER_URI),
            self.environment.get(KEY_K8S_MASTER_URI_ENV),
            self.environment.get(KEY_K8S_MASTER_URI_ENV_DEPRECATED))

    def _k8s_ca_base64(self):
        return PriorityEvaluator._first(
            self.cli_arguments.get(KEY_K8S_CA_BASE64),
            self.context_arguments.get(KEY_K8S_CA_BASE64),
            self.environment.get(KEY_K8S_CA_BASE64_ENV),
            self.environment.get(KEY_K8S_CA_BASE64_URI_ENV_DEPRECATED))

    def _k8s_token(self):
        return PriorityEvaluator._first(
            self.cli_arguments.get(KEY_K8S_TOKEN),
            self.context_arguments.get(KEY_K8S_TOKEN),
            self.environment.get(KEY_K8S_TOKEN_ENV))

    def _k8s_handle_debug(self):
        return PriorityEvaluator._first(
            self.cli_arguments.get(KEY_K8S_HANDLE_DEBUG),
            self.context_arguments.get(KEY_K8S_HANDLE_DEBUG) in [True, 'true', 'True'])

    @staticmethod
    def _first(*arguments):
        if not arguments:
            return None

        for argument in arguments:
            if not argument:
                continue

            return argument

        return None


def _process_variable(variable):
    matches = INCLUDE_RE.match(variable)

    if matches:
        return load_yaml(matches.groupdict().get('file'))

    try:
        return re.sub(CUSTOM_ENV_RE, lambda m: os.environ[m.group(1)], variable)

    except KeyError as err:
        log.debug('Environment variable "{}" is not set'.format(err.args[0]))
        if settings.GET_ENVIRON_STRICT:
            raise RuntimeError('Environment variable "{}" is not set'.format(err.args[0]))

    return re.sub(CUSTOM_ENV_RE, lambda m: os.environ.get(m.group(1), ''), variable)


def _update_single_variable(value, include_history):
    if value in include_history:
        raise RuntimeError('Infinite include loop')

    local_history = copy.copy(include_history)
    local_history.append(value)

    return _update_context_recursively(_process_variable(value), local_history)


def _update_context_recursively(context, include_history=[]):
    if isinstance(context, dict):
        output = {}
        for key, value in context.items():
            if isinstance(value, str):
                output[key] = _update_single_variable(value, include_history)
            else:
                output[key] = _update_context_recursively(value)
        return output
    elif isinstance(context, list):
        output = []
        for value in context:
            if isinstance(value, str):
                output.append(_update_single_variable(value, include_history))
            else:
                output.append(_update_context_recursively(value))
        return output
    else:
        return context


def load_context_section(section):
    if not section:
        raise RuntimeError('Empty section specification is not allowed')

    if section == settings.COMMON_SECTION_NAME:
        raise RuntimeError('Section "{}" is not intended to deploy'.format(settings.COMMON_SECTION_NAME))

    context = load_yaml(settings.CONFIG_FILE)

    if context is None:
        raise RuntimeError('Config file "{}" is empty'.format(settings.CONFIG_FILE))

    if section not in context:
        raise RuntimeError('Section "{}" not found in config file "{}"'.format(section, settings.CONFIG_FILE))

    # delete all sections except common and used section
    context.setdefault(settings.COMMON_SECTION_NAME, {})
    context = {key: context[key] for key in [settings.COMMON_SECTION_NAME, section]}
    context = _update_context_recursively(context)

    if section and section in context:
        context = merge(context[settings.COMMON_SECTION_NAME], context[section])

    if 'templates' not in context and 'kubectl' not in context:
        raise RuntimeError(
            'Section "templates" or "kubectl" not found in config file "{}"'.format(settings.CONFIG_FILE))

    validate_dashes(context)
    return context


def get_all_nested_keys(result, d):
    for key, value in d.items():
        result.append(key)
        if isinstance(d[key], dict):
            get_all_nested_keys(result, d[key])

    return result


def get_vars_with_dashes(vars_list):
    return [var_name for var_name in vars_list if '-' in var_name]


def validate_dashes(context):
    all_keys = get_all_nested_keys([], context)
    dashes = get_vars_with_dashes(all_keys)
    if len(dashes) != 0:
        raise RuntimeError('Variable names should never include dashes, '
                           'check your vars, please: {}'.format(', '.join(sorted(dashes))))



import logging

from kubernetes import client
from kubernetes.client.rest import ApiException

from k8s_handle import settings
from k8s_handle.exceptions import ProvisioningError
from k8s_handle.transforms import add_indent, split_str_by_capital_letters
from .api_extensions import ResourcesAPI
from .mocks import K8sClientMock

log = logging.getLogger(__name__)


class Adapter:
    api_versions = {
        'apps/v1beta1': client.AppsV1beta1Api,
        'v1': client.CoreV1Api,
        'extensions/v1beta1': client.ExtensionsV1beta1Api,
        'batch/v1': client.BatchV1Api,
        'batch/v2alpha1': client.BatchV2alpha1Api,
        'batch/v1beta1': client.BatchV1beta1Api,
        'policy/v1beta1': client.PolicyV1beta1Api,
        'storage.k8s.io/v1': client.StorageV1Api,
        'apps/v1': client.AppsV1Api,
        'autoscaling/v1': client.AutoscalingV1Api,
        'rbac.authorization.k8s.io/v1': client.RbacAuthorizationV1Api,
        'scheduling.k8s.io/v1alpha1': client.SchedulingV1alpha1Api,
        'scheduling.k8s.io/v1beta1': client.SchedulingV1beta1Api,
        'networking.k8s.io/v1': client.NetworkingV1Api,
        'apiextensions.k8s.io/v1beta1': client.ApiextensionsV1beta1Api,
    }
    kinds_builtin = [
        'ConfigMap', 'CronJob', 'DaemonSet', 'Deployment', 'Endpoints',
        'Ingress', 'Job', 'Namespace', 'PodDisruptionBudget', 'ResourceQuota',
        'Secret', 'Service', 'ServiceAccount', 'StatefulSet', 'StorageClass',
        'PersistentVolume', 'PersistentVolumeClaim', 'HorizontalPodAutoscaler',
        'Role', 'RoleBinding', 'ClusterRole', 'ClusterRoleBinding', 'CustomResourceDefinition',
        'PriorityClass', 'PodSecurityPolicy', 'LimitRange', 'NetworkPolicy'
    ]

    def __init__(self, spec):
        self.body = spec
        self.kind = spec.get('kind', "")
        self.name = spec.get('metadata', {}).get('name')
        self.namespace = spec.get('metadata', {}).get('namespace', "") or settings.K8S_NAMESPACE

    @staticmethod
    def get_instance(spec, api_custom_objects=None, api_resources=None):
        # due to https://github.com/kubernetes-client/python/issues/387
        if spec.get('kind') in Adapter.kinds_builtin:
            if spec.get('apiVersion') == 'test/test':
                return AdapterBuiltinKind(spec, K8sClientMock(spec.get('metadata', {}).get('name')))

            api = Adapter.api_versions.get(spec.get('apiVersion'))

            if not api:
                return None

            return AdapterBuiltinKind(spec, api())

        api_custom_objects = api_custom_objects or client.CustomObjectsApi()
        api_resources = api_resources or ResourcesAPI()
        return AdapterCustomKind(spec, api_custom_objects, api_resources)


class AdapterBuiltinKind(Adapter):
    def __init__(self, spec, api=None):
        super().__init__(spec)
        self.kind = split_str_by_capital_letters(spec['kind'])
        self.replicas = spec.get('spec', {}).get('replicas')
        self.api = api

    def get(self):
        try:
            if hasattr(self.api, "read_namespaced_{}".format(self.kind)):
                response = getattr(self.api, 'read_namespaced_{}'.format(self.kind))(
                    self.name, namespace=self.namespace)
            else:
                response = getattr(self.api, 'read_{}'.format(self.kind))(self.name)
        except ApiException as e:
            if e.reason == 'Not Found':
                return None
            log.error('Exception when calling "read_namespaced_{}": {}'.format(self.kind, add_indent(e.body)))
            raise ProvisioningError(e)

        return response

    def get_pods_by_selector(self, label_selector):
        try:
            if not isinstance(self.api, K8sClientMock):
                self.api = client.CoreV1Api()

            return self.api.list_namespaced_pod(
                namespace=self.namespace, label_selector='job-name={}'.format(label_selector))

        except ApiException as e:
            log.error('Exception when calling CoreV1Api->list_namespaced_pod: {}', e)
            raise e

    def read_pod_status(self, name):
        try:
            if not isinstance(self.api, K8sClientMock):
                self.api = client.CoreV1Api()

            return self.api.read_namespaced_pod_status(name, namespace=self.namespace)
        except ApiException as e:
            log.error('Exception when calling CoreV1Api->read_namespaced_pod_status: {}', e)
            raise e

    def read_pod_logs(self, name, container):
        log.info('Read logs for pod "{}", container "{}"'.format(name, container))
        try:
            if not isinstance(self.api, K8sClientMock):
                self.api = client.CoreV1Api()
            if settings.COUNT_LOG_LINES:
                return self.api.read_namespaced_pod_log(name, namespace=self.namespace, timestamps=True,
                                                        tail_lines=settings.COUNT_LOG_LINES, container=container)
            return self.api.read_namespaced_pod_log(name, namespace=self.namespace, timestamps=True,
                                                    container=container)
        except ApiException as e:
            log.error('Exception when calling CoreV1Api->read_namespaced_pod_log: {}', e)
            raise e

    def create(self):
        try:
            if hasattr(self.api, "create_namespaced_{}".format(self.kind)):
                return getattr(self.api, 'create_namespaced_{}'.format(self.kind))(
                    body=self.body, namespace=self.namespace)

            return getattr(self.api, 'create_{}'.format(self.kind))(body=self.body)
        except ApiException as e:
            log.error('Exception when calling "create_namespaced_{}": {}'.format(self.kind, add_indent(e.body)))
            raise ProvisioningError(e)
        except ValueError as e:
            log.error(e)
            # WORKAROUND https://github.com/kubernetes-client/python/issues/466
            # also https://github.com/kubernetes-client/gen/issues/52
            if self.kind not in ['pod_disruption_budget', 'custom_resource_definition']:
                raise e

    def replace(self, parameters):
        try:
            if self.kind in ['custom_resource_definition']:
                self.body['metadata']['resourceVersion'] = parameters['resourceVersion']
                return self.api.replace_custom_resource_definition(
                    self.name, self.body,
                )

            if self.kind in ['service', 'service_account']:
                if 'spec' in self.body:
                    self.body['spec']['ports'] = parameters.get('ports')

                return getattr(self.api, 'patch_namespaced_{}'.format(self.kind))(
                    name=self.name, body=self.body, namespace=self.namespace
                )

            if hasattr(self.api, "replace_namespaced_{}".format(self.kind)):
                return getattr(self.api, 'replace_namespaced_{}'.format(self.kind))(
                    name=self.name, body=self.body, namespace=self.namespace)

            return getattr(self.api, 'replace_{}'.format(self.kind))(
                name=self.name, body=self.body)
        except ApiException as e:
            log.error('Exception when calling "replace_namespaced_{}": {}'.format(self.kind, add_indent(e.body)))
            raise ProvisioningError(e)

    def delete(self):
        try:
            if hasattr(self.api, "delete_namespaced_{}".format(self.kind)):
                return getattr(self.api, 'delete_namespaced_{}'.format(self.kind))(
                    name=self.name, body=client.V1DeleteOptions(propagation_policy='Foreground'),
                    namespace=self.namespace)

            return getattr(self.api, 'delete_{}'.format(self.kind))(
                name=self.name, body=client.V1DeleteOptions(propagation_policy='Foreground'))
        except ApiException as e:
            if e.reason == 'Not Found':
                return None
            log.error('Exception when calling "delete_namespaced_{}": {}'.format(self.kind, add_indent(e.body)))
            raise ProvisioningError(e)


class AdapterCustomKind(Adapter):
    def __init__(self, spec, api_custom_objects, api_resources):
        super().__init__(spec)
        self.api = api_custom_objects
        self.api_resources = api_resources
        self.plural = None

        try:
            api_version_splitted = spec.get('apiVersion').split('/', 1)
            self.group = api_version_splitted[0]
            self.version = api_version_splitted[1]
        except (IndexError, AttributeError):
            self.group = None
            self.version = None

        resources_list = self.api_resources.list_api_resource_arbitrary(self.group, self.version)

        if not resources_list:
            return

        for resource in resources_list.resources:
            if resource.kind != self.kind:
                continue

            self.plural = resource.name

            if not resource.namespaced:
                self.namespace = ""

            break

    def get(self):
        self._validate()

        try:
            if self.namespace:
                return self.api.get_namespaced_custom_object(
                    self.group, self.version, self.namespace, self.plural, self.name
                )

            return self.api.get_cluster_custom_object(self.group, self.version, self.plural, self.name)

        except ApiException as e:
            if e.reason == 'Not Found':
                return None

            log.error('{}'.format(add_indent(e.body)))
            raise ProvisioningError(e)

    def create(self):
        self._validate()

        try:
            if self.namespace:
                return self.api.create_namespaced_custom_object(
                    self.group, self.version, self.namespace, self.plural, self.body
                )

            return self.api.create_cluster_custom_object(self.group, self.version, self.plural, self.body)

        except ApiException as e:
            log.error('{}'.format(add_indent(e.body)))
            raise ProvisioningError(e)

    def delete(self):
        self._validate()

        try:
            if self.namespace:
                return self.api.delete_namespaced_custom_object(
                    self.group, self.version, self.namespace, self.plural, self.name,
                    client.V1DeleteOptions(propagation_policy='Foreground')
                )

            return self.api.delete_cluster_custom_object(
                self.group, self.version, self.plural, self.name,
                client.V1DeleteOptions(propagation_policy='Foreground')
            )

        except ApiException as e:
            if e.reason == 'Not Found':
                return None

            log.error(
                '{}'.format(add_indent(e.body)))
            raise ProvisioningError(e)

    def replace(self, _):
        self._validate()

        try:
            if self.namespace:
                return self.api.patch_namespaced_custom_object(
                    self.group, self.version, self.namespace, self.plural, self.name, self.body
                )

            return self.api.patch_cluster_custom_object(
                self.group, self.version, self.plural, self.name, self.body
            )
        except ApiException as e:
            log.error('{}'.format(add_indent(e.body)))
            raise ProvisioningError(e)

    def _validate(self):
        if not self.plural:
            raise RuntimeError("No valid plural name of resource definition discovered")

        if not self.group:
            raise RuntimeError("No valid resource definition group discovered")

        if not self.version:
            raise RuntimeError("No valid version of resource definition supplied")



import unittest

from k8s_handle.exceptions import DeprecationError
from k8s_handle.k8s.deprecation_checker import ApiDeprecationChecker


class TestApiDeprecationChecker(unittest.TestCase):

    def test_version_not_in_list(self):
        checker = ApiDeprecationChecker("1.9.7")
        checker.deprecated_versions = {
            "test/v1": {
                "since": "1.8.0",
                "until": "1.10.0",
                "resources": [
                    "Deployment",
                ],
            }
        }
        self.assertFalse(checker._is_deprecated("test/v2", "Deployment"))

    def test_kind_not_in_list(self):
        checker = ApiDeprecationChecker("1.9.7")
        checker.deprecated_versions = {
            "test/v1": {
                "since": "1.8.0",
                "until": "1.10.0",
                "resources": [
                    "Deployment",
                ],
            }
        }
        self.assertFalse(checker._is_deprecated("test/v1", "StatefulSet"))

    def test_version_not_deprecated_yet(self):
        checker = ApiDeprecationChecker("1.7.9")
        checker.deprecated_versions = {
            "test/v1": {
                "since": "1.8.0",
                "until": "1.10.0",
                "resources": [
                    "Deployment",
                ],
            }
        }
        self.assertFalse(checker._is_deprecated("test/v1", "Deployment"))

    def test_version_is_deprecated_equal(self):
        checker = ApiDeprecationChecker("1.8.0")
        checker.deprecated_versions = {
            "test/v1": {
                "since": "1.8.0",
                "until": "1.10.0",
                "resources": [
                    "Deployment",
                ],
            }
        }
        self.assertTrue(checker._is_deprecated("test/v1", "Deployment"))

    def test_version_is_deprecated(self):
        checker = ApiDeprecationChecker("1.9.9")
        checker.deprecated_versions = {
            "test/v1": {
                "since": "1.8.0",
                "until": "1.10.0",
                "resources": [
                    "Deployment",
                ],
            }
        }
        self.assertTrue(checker._is_deprecated("test/v1", "Deployment"))

    def test_version_is_unsupported_equal(self):
        checker = ApiDeprecationChecker("1.10.0")
        checker.deprecated_versions = {
            "test/v1": {
                "since": "1.8.0",
                "until": "1.10.0",
                "resources": [
                    "Deployment",
                ],
            }
        }
        with self.assertRaises(DeprecationError):
            checker._is_deprecated("test/v1", "Deployment")

    def test_version_is_unsupported(self):
        checker = ApiDeprecationChecker("1.10.6")
        checker.deprecated_versions = {
            "test/v1": {
                "since": "1.8.0",
                "until": "1.10.0",
                "resources": [
                    "Deployment",
                ],
            }
        }
        with self.assertRaises(DeprecationError):
            checker._is_deprecated("test/v1", "Deployment")

    def test_version_no_until(self):
        checker = ApiDeprecationChecker("1.10.6")
        checker.deprecated_versions = {
            "test/v1": {
                "since": "1.8.0",
                "until": "",
                "resources": [
                    "Deployment",
                ],
            }
        }
        self.assertTrue(checker._is_deprecated("test/v1", "Deployment"))
